#include "rs_denoise.h"

static void error(char *name);

static void bilinear_decim(
  struct pxm_img *out_img,
  struct pxm_img *img,
  int decim_factor);

static void scanner_decim(
  struct pxm_img *out_img,
  struct pxm_img *img,
  int decim_factor);

static void near_neighbor_decim(
  struct pxm_img *out_img,
  struct pxm_img *img,
  int decim_factor);

int main(argc, argv) 
int  argc;
char *argv[];
{
  FILE 			 *fp;
  
  char  	         fname[1006], decim_mode[100];
  char          	 out_format[1000], in_format[1000];
  
  int 			 k, decim_factor, num_of_images, num_of_zeros;
/*  
  double 		 h_quotient, w_quotient;
*/
  time_t 	         first, second;
  unsigned char          preprocess_flag;
  struct pxm_img 	 in_img, lpf_img, out_img, *img;
  
  /* check for correct number of arguments */
  if(argc != 7) error(argv[0]);
  
  /* read decimation factor */
  sscanf(argv[3],"%d",&decim_factor);
  
  /* read number of images */
  sscanf(argv[4],"%d",&num_of_images);
  if ( num_of_images <= 0 ) {
    fprintf(stderr,"\nError: number of images must be positive\n");
    exit(1);
  }

  /* read decimation mode */
  sscanf(argv[5],"%s",decim_mode);

  /* read the preprocess_flag */
  preprocess_flag = (unsigned char)atoi(argv[6]);
  if(preprocess_flag != 0 && preprocess_flag != 1) {
    fprintf(stderr, "\nError in main(): preprocess_flag should ");
    fprintf(stderr, "be either 0 or 1\n");
    exit(1);
  }
  
  /* set variable to do timing of algorithm */
  first = time(NULL);
  fprintf(stdout,"\nRun decimation ...");

  /* string for formatting image names */
  if (num_of_images > 1) {
    num_of_zeros = 1 + (int) log10 ( (double) (num_of_images - 1));
    sprintf (in_format, "%s%%0%dd", argv[1], num_of_zeros);
    sprintf (out_format, "%s%%0%dd", argv[2], num_of_zeros);
  }
  
  for (k = 0; k < num_of_images; k++) {
    
    /* get file name */
    if (num_of_images == 1) sprintf(fname, "%s", argv[1]);
    if (num_of_images > 1) sprintf(fname, in_format, k);
    
    /* read image */
    if((fp=fopen(fname,"r"))==NULL) {
      fprintf(stderr,"\nError: can't open image %s\n",fname);
      exit(1);
    }
    if (read_pxm (fp, &in_img)) {
      fprintf(stderr,"\nError: can't read pxm file\n");
      exit(1);
    }
    fclose(fp);
    
    /* check for PGM/PPM file type */
    if ( (in_img.pxm_type != 'g') && (in_img.pxm_type != 'p') ) {
      fprintf(stderr,"\nError: input image %s not a PGM/PPM file.\n",fname);
      exit(1);
    }
    
    /* allocate output image */
    if(in_img.pxm_type == 'g')
      get_pxm(&out_img, in_img.height / decim_factor, 
        in_img.width / decim_factor, 'g');
    if(in_img.pxm_type == 'p')
      get_pxm(&out_img, in_img.height / decim_factor, 
  	in_img.width / decim_factor, 'p');
  
    /* Make sure decim_factor divides into image size */
/*
    h_quotient = (double) in_img.height / (double) decim_factor;
    w_quotient = (double) in_img.width / (double) decim_factor;
    if ( ( h_quotient - (int) h_quotient > 0 ) ||
	 ( w_quotient - (int) w_quotient > 0 ) ) {
      fprintf ( stderr, "Warning:  decimation factor does" );
      fprintf ( stderr, "\nnot divide size of image %s\n",fname );
    }
*/       
    if(preprocess_flag == 1){
      /* turn-on preprocessing */
      apply_gauss_filter( &in_img, &lpf_img, 1.0, 1);
      img = &lpf_img;
    }
    else {
      /* turn-off preprocessing */
      img = &in_img;
    }
     
    if(!strcmp("bilinear", decim_mode))
      bilinear_decim(&out_img, img, decim_factor);

    else if(!strcmp("scanner", decim_mode))
      scanner_decim(&out_img, img, decim_factor);

    else if(!strcmp("nearest", decim_mode)) 
      near_neighbor_decim(&out_img, img, decim_factor);

    else {
      fprintf(stderr,"\nInvalid decimation mode: Available modes are \n");
      fprintf(stderr,"'bilinear', 'nearest', and 'scanner' \n");
      exit(1);
    }

    /* show the run time */
    second = time(NULL);
    fprintf(stdout,"\nFinished decimation, now writing results.");
    fprintf(stdout,"\nRun time        :  %f sec.\n",difftime(second,first));

    /* write Image */
    if ( num_of_images == 1 ) sprintf ( fname, "%s", argv[2] );
    if ( num_of_images > 1 ) sprintf ( fname, out_format, k );
    if((fp=fopen(fname,"w")) == NULL) {
      fprintf(stderr,"\nError: can't write to data file\n");
      exit(1);
    }
    if ( write_pxm(fp, &out_img) ){
      fprintf(stderr,"\nError: can't write to pxm file\n");
      exit(1);
    }
    fclose(fp);
    /*
      free_pxm ( &img );
      free_pxm ( &out_img );
    */
  }
  
return 0;
}

static void error(char *name)
{
  printf("Usage:  %s  input_image_basename   output_image_basename  ",name);
  printf("decimation_factor   #_of_images\n");
  exit(1);
}

static void bilinear_decim(
  struct pxm_img *out_img,
  struct pxm_img *img,
  int decim_factor)
{   
  int 			 i, j, m, n, temp;
  int			 tempR, tempG, tempB;
  double 		 area;


  area = (double) (decim_factor * decim_factor);
  for(i=0; i<out_img->height; i++)
    for(j=0; j<out_img->width; j++) {
      if ( out_img->pxm_type == 'g' ) {
	temp = 0;
	for(m=0; m<decim_factor; m++)
	  for(n=0; n<decim_factor; n++) {
	    temp += img->mono[decim_factor*i+m][decim_factor*j+n];
	      }
	out_img->mono[i][j] = (unsigned char) (0.5+(temp/area));
      }
      if ( out_img->pxm_type == 'p' ) {
	tempR=0;tempG=0;tempB=0;
	for(m=0; m<decim_factor; m++)
	  for(n=0; n<decim_factor; n++) {
	    tempR += img->color[0][decim_factor*i+m][decim_factor*j+n];
	    tempG += img->color[1][decim_factor*i+m][decim_factor*j+n];
	    tempB += img->color[2][decim_factor*i+m][decim_factor*j+n];
	  }
	out_img->color[0][i][j] = (unsigned char) (0.5+(tempR/area));
	out_img->color[1][i][j] = (unsigned char) (0.5+(tempG/area));
	out_img->color[2][i][j] = (unsigned char) (0.5+(tempB/area));
      }
    }
}

static void scanner_decim(
  struct pxm_img *out_img,
  struct pxm_img *img,
  int decim_factor)

{
  int 			 i, j, n, temp;
  int			 tempR, tempG, tempB;
  double 		 area;

  area = (double) (decim_factor);
  for(i=0; i<out_img->height; i++)
    for(j=0; j<out_img->width; j++) {
      if ( out_img->pxm_type == 'g' ) {
	temp = 0;
	for(n=0; n<decim_factor; n++)
	  temp += img->mono[decim_factor*i][decim_factor*j+n];
	out_img->mono[i][j] = (unsigned char) (0.5+(temp/area));
      }
      if ( out_img->pxm_type == 'p' ) {
	tempR=0;tempG=0;tempB=0;
	for(n=0; n<decim_factor; n++) {
	  tempR += img->color[0][decim_factor*i][decim_factor*j+n];
	  tempG += img->color[1][decim_factor*i][decim_factor*j+n];
	  tempB += img->color[2][decim_factor*i][decim_factor*j+n];
	}
	out_img->color[0][i][j] = (unsigned char) (0.5+(tempR/area));
	out_img->color[1][i][j] = (unsigned char) (0.5+(tempG/area));
	out_img->color[2][i][j] = (unsigned char) (0.5+(tempB/area));
      }
    }
}

static void  near_neighbor_decim(
  struct pxm_img *out_img,
  struct pxm_img *img,
  int decim_factor)

{

  int 			 i, j;

  for(i=0; i<out_img->height; i++)
    for(j=0; j<out_img->width; j++) {
      if( out_img->pxm_type == 'g' ) 
	out_img->mono[i][j] = img->mono[decim_factor*i][decim_factor*j];
      
      if(out_img->pxm_type == 'p') {
	out_img->color[0][i][j] = img->color[0][decim_factor*i][decim_factor*j];
	out_img->color[1][i][j] = img->color[1][decim_factor*i][decim_factor*j];
	out_img->color[2][i][j] = img->color[2][decim_factor*i][decim_factor*j];
      }
    }
}
