#include "rs_denoise.h"

static void error(char *name);

static void
apply_filter(
	     float *prob,
	     float ***filter,
	     float **joint_means,
	     int num_filts,
	     int *filt_indices,
	     float *in_vec,
	     int in_length,
	     float *out_vec,
	     int out_length);

int main(int argc, char **argv) 
{
  struct pxm_img     scan_img, lpf_img, imgh, *img_ptr;
  struct fOptSigSet  *OS;

  int                filt_count, i, j, interp_factor;
  int                W, interp_clust_length;
  int                P, filt_length, high_res_length;
  int                max, num_filts;
  int                *filt_indices;
  int		     scan_res;

  float              avg_filts, delta_factor, clust_vec_power;
  float              *prob, *clust_vect, *in_vec, *out_vec;
  float              **joint_means, ***filter;

  time_t 	     first, second;

  unsigned char      preprocess_flag;

  void 		     (*get_clust_vector) (unsigned char **, float*, float, int, 
			int, int, int, int);

#ifdef SEGMENTATION_MAP
  struct pxm_img     class_img;
#endif

  /* check for correct number of arguments */
  if (argc != 7) error (argv[0]);

  /* read the raw scanned image */
  read_pxm_img(argv[1], &scan_img);
  if(scan_img.pxm_type !='p' && scan_img.pxm_type !='g'){
    printf("Error in main(): \n %s should be type p or g\n",argv[1]);
    exit(1);
  }
  if(scan_img.pxm_type == 'p') convertPPMtoPGM(&scan_img);

  /* read the preprocess_flag */
  preprocess_flag = (unsigned char)atoi(argv[6]);
  if(preprocess_flag != 0 && preprocess_flag != 1) {
    fprintf(stderr, "\nError in main(): preprocess_flag should ");
    fprintf(stderr, "be either 0 or 1\n");
    exit(1);
  }

  /* read classifier and filter parameters */
  read_classifier_filter (argv[3], &OS, &interp_factor, &W, &P, 
    &interp_clust_length, &filt_length, &high_res_length, &clust_vec_power, 
    &joint_means, &filter) ;

  /* set delta_factor for determining */
  /* classes to use in filtering      */
  delta_factor = (float)(atof(argv[4]));
  if ( delta_factor < 0.0 ) {
    fprintf(stderr,"\nerror: delta_factor must be >= 0\n");
    exit(1);
  }

  /* determine the scanner resolution */
  scan_res = atoi(argv[5]);
  if(scan_res != 300 && scan_res != 600) {
    fprintf(stderr,"\nerror: scan_res should be either 300 or 600\n");
    exit(1);
  }

  /* set variable to do timing of algorithm */
  first = time(NULL);
  fprintf(stdout,"\nRun RS-based denoising ...");

  if(preprocess_flag == 1){
    /* turn-on preprocessing */
    apply_gauss_filter( &scan_img, &lpf_img, 0.6, 1);
    img_ptr = &lpf_img;
  }
  else {
    /* turn-off preprocessing */
    img_ptr = &scan_img;
  }

  /* allocate scratch memory */
  clust_vect   = (float *)get_spc(interp_clust_length, sizeof(float));
  in_vec       = (float *)get_spc(filt_length, sizeof(float));
  out_vec      = (float *)get_spc(high_res_length, sizeof(float));
  prob         = (float *)get_spc(OS->OptClassSig->nsubclasses, sizeof(float));
  filt_indices = (int *)get_spc(OS->OptClassSig->nsubclasses, sizeof(int));

#ifdef SEGMENTATION_MAP
  /* allocate space for the class image */
  get_pxm (&class_img, scan_img.height, scan_img.width, 'g');
#endif

  /* cluster vector for rs-enhancement */
  get_clust_vector = &get_den_clust_vector;

  /* apply RS-based denoising/enhancement  */
  filt_count=0;

  /* allocate space for high resolution image */
  get_pxm(&imgh, interp_factor * scan_img.height,
	  interp_factor * scan_img.width, 'g');

  for(i = P; i < scan_img.height - P; i++)
    for(j = P; j < scan_img.width - P; j++) {

      /* get cluster vector from luminance componenent */
      get_clust_vector(img_ptr->mono, clust_vect, clust_vec_power, W, i, j, 
        scan_img.height, scan_img.width);

      /* choose which classes to use in filtering     */
      num_filts = choose_classes( filt_indices, delta_factor, prob, clust_vect,
        OS->OptClassSig, interp_clust_length,&max );

      filt_count += num_filts;

#ifdef SEGMENTATION_MAP   
      class_img.mono[i][j] = max;
#endif

      /* red pixels */
      GetLowResPixels(scan_img.mono, in_vec, P, i, j, scan_img.height, 
        scan_img.width);
      apply_filter(prob, filter, joint_means, num_filts, filt_indices, 
        in_vec, filt_length, out_vec, high_res_length);
      put_high_res_pixels(imgh.mono, scan_img.mono, interp_factor, 
	i, j, out_vec);

#ifdef SEGMENTATION_MAP
    class_img.mono[i][j]=(unsigned char) max;
#endif
    }

  /* take care of the boundary pixels */
  for(i = 0; i < interp_factor * P; i++)
    for(j = interp_factor * P; j < imgh.width - interp_factor * P; j++){
      /* top */
      imgh.mono[i][j] = imgh.mono[interp_factor * P][j];
      /* bottom */
      imgh.mono[i + imgh.height - interp_factor * P][j] =
        imgh.mono[i + imgh.height -interp_factor * P - 1][j];
  }

  for(i = 0; i < imgh.height; i++)
    for(j = 0; j < interp_factor * P; j++){
      /* left */
      imgh.mono[i][j] = imgh.mono[i][interp_factor * P];
      /* right */
      imgh.mono[i][j + imgh.width - interp_factor * P] =
        imgh.mono[i][j + imgh.width - interp_factor * P - 1];
  }

  avg_filts = (float)(filt_count) / (float)((scan_img.height)*(scan_img.width));

  /* show the run time */
  second = time(NULL);
  fprintf(stdout,"\nFinished RS-based denoising, now writing results.");
  fprintf(stdout,"\nAvg. num filters:  %f   ",avg_filts);
  fprintf(stdout,"\nRun time        :  %f sec.\n",difftime(second,first));

  /* write interpolated and class images */
  write_pxm_img(argv[2], &imgh);

#ifdef SEGMENTATION_MAP  
  write_pxm_img("class_img", &class_img);
#endif

  return(0);
}

static void error(char *name)
{
  printf("\nusage: %s     scan_img.pxm  ",name);
  printf("\n\t\t\tdescr_img.pxm");
  printf("\n\t\t\tfilter_file");
  printf("\n\t\t\tdelta_factor must be >= 0:");
  printf("\n\t\t\t  = 0  :  Efficient RS - 1 filter used per pixel");
  printf("\n\t\t\t  > 12 :  Full RS - all filters used per pixel");
  printf("\n\t\t\tscan_resolution must be 300 or 600");
  printf("\n\t\t\tpreprocess_flag must be 0 or 1\n\n");

  exit(1);
}

static void
apply_filter(
	     float *prob,
	     float ***filter,
	     float **joint_means,
	     int num_filts,
	     int *filt_indices,
	     float *in_vec,
	     int in_length,
	     float *out_vec,
	     int out_length)
{
  static int first=1;
  static float *buff;
  float *buff_pt,*buff_pt_max,*filt_pt;
  float *high_res_mean_pt,*low_res_mean_pt;
  float *out_vec_pt,*in_vec_pt,*in_vec_pt_max;
  float *prob_pt,*prob_pt_max;
  float diff;
  int *filt_indices_pt;
  int i,index;

  if(first) {
    buff = (float *)get_spc(out_length,sizeof(float));
    first = 0;
  }

  for(i=0; i<out_length; i++) out_vec[i]=0;

  /* set max pointer values */
  prob_pt_max=prob+num_filts;
  in_vec_pt_max=in_vec+in_length;
  buff_pt_max=buff+out_length;

  /* THIS LOOP DOMINATES THE COMPUTATION */
  /* apply each of the "num_filt" filters */
  /* indexed by the "filt_indices" list   */
  for ( prob_pt=prob,filt_indices_pt=filt_indices; 
	prob_pt<prob_pt_max; 
	prob_pt++,filt_indices_pt++ ) {

    /* clear buffer */
    for (buff_pt=buff;buff_pt<buff_pt_max;buff_pt++) *buff_pt=0;

    /* get filter index */
    index = *filt_indices_pt;

    /* matrix multiply */
    filt_pt = *filter[index];
    for ( in_vec_pt=in_vec,low_res_mean_pt=joint_means[index]+out_length;
	  in_vec_pt<in_vec_pt_max; 
	  in_vec_pt++,low_res_mean_pt++ ) {
      diff = *in_vec_pt - *low_res_mean_pt;  

      for ( buff_pt=buff;
	    buff_pt<buff_pt_max;
	    buff_pt++,filt_pt++ ) {
	*buff_pt += *filt_pt * diff;
      }
    }

    /* add mean vector and scale by probability */
    for ( buff_pt=buff,out_vec_pt=out_vec,high_res_mean_pt=joint_means[index];
	  buff_pt<buff_pt_max;
	  buff_pt++,out_vec_pt++,high_res_mean_pt++ ) 
      *out_vec_pt += *prob_pt * ( (*buff_pt)  +  (*high_res_mean_pt) );

  }
}




  
