#include "susan_smooth.h"

static void procCmdLine ( int argc, char **argv, 
			  char **in_fname1,
			  char **in_fname2,
 			  char **out_fname,
                          int  *bt, 
			  int  *form,
			  int  *mask_size,
			  float *dt,
			  unsigned char *tone_flag );

static void error( char *name );

int main ( int argc, char **argv ) 
{
  FILE *fp;

  int bt;
  int form;
  int mask_size;
  int k;

  char *in_fname1;
  char *in_fname2;
  char *out_fname;

  float *bp, *dp; 
  float dt;
  float **lin_in, **lin_out;

  time_t first, second;

  unsigned char tone_flag;

  struct pxm_img in_img1, in_img2, out_img;
  
  /* read command line */
  procCmdLine( argc, argv, &in_fname1,&in_fname2, &out_fname, 
	       &bt, &form, &mask_size, &dt, &tone_flag ); 

   /* read the denoised image */
  if((fp=fopen(in_fname1,"r"))==NULL) {
    fprintf(stderr,"\nError: can't open image %s\n",in_fname1);
    exit(1);
  }
  if (read_pxm (fp, &in_img1)) {
    fprintf(stderr,"\nError: can't read pxm file\n");
    exit(1);
  }
  fclose(fp);

  if(in_img1.pxm_type == 'g') convertPGMtoPPM(&in_img1);
  
  /* read the scanned image */
  if((fp=fopen(in_fname2,"r"))==NULL) {
    fprintf(stderr,"\nError: can't open image %s\n",in_fname2);
    exit(1);
  }
  if (read_pxm (fp, &in_img2)) {
    fprintf(stderr,"\nError: can't read pxm file\n");
    exit(1);
  }
  fclose(fp);

  if(in_img2.pxm_type == 'g') convertPGMtoPPM(&in_img2);

  /* Set variable to do timing of algorithm */
  first = time(NULL);
  fprintf(stdout,"\nRun Susan filtering ...");

  /* check the input image dimensions */
  if( in_img1.height > in_img2.height ||
      in_img1.width > in_img2.width )
    {
      printf( "dimensions of input images do not match\n");
      exit(1);
    }

  /* allocate memory */
  lin_in = ( float ** )
    get_img( in_img1.width, in_img1.height, sizeof(float) );

  lin_out = ( float ** )
    get_img( in_img1.width, in_img1.height, sizeof(float) );  
  
  get_pxm (&out_img, in_img1.height, in_img1.width, 'p');

  /* run susan filtering in linear domain */
  setup_brightness_lut(&bp, bt, form);
  dp = make_gauss_filt( mask_size, dt, in_img1.width, in_img1.height );

  for(k = 0; k < 3; k++){

    /* gamma uncorrect the input scanned image */
    inv_scanner_fun(in_img2.color[k], lin_in, in_img1.height, in_img1.width,
      tone_flag);

    /* apply filter */
    susan_smooth(in_img1.color[k], lin_in, &lin_out, bp, dp, mask_size, 
      in_img1.width, in_img1.height);
    
   /* gamma correct the output linear image */
    inv_printer_fun(out_img.color[k], lin_out, out_img.height, out_img.width, 
      tone_flag);
  }

  /* show the run time */
  second = time(NULL);
  printf("\nFinished Susan filtering, now writing results.");
  printf("\nRun time         : %f sec.\n",difftime(second,first));

  /*write the output image*/
  if((fp=fopen(out_fname,"w"))==NULL) {
    fprintf(stderr,"\nError: can't write to data file\n");
    exit(1);
  }
  if ( write_pxm(fp, &out_img) ) {
    fprintf(stderr,"\nError: can't write to pxm file\n");
    exit(1);
  }
  fclose(fp);

  return 0;
}
                
static void procCmdLine ( int argc, char **argv, 
			  char **in_fname1,
			  char **in_fname2,
 			  char **out_fname,
                          int  *bt, 
			  int  *form,
			  int  *mask_size,
			  float *dt,
			  unsigned char *tone_flag )
{
  /* check for correct number of arguments */
  if(argc!=9)error(argv[0]);

  *in_fname1 = argv[1];
  *in_fname2 = argv[2];
  *out_fname = argv[3];

  *bt = atoi(argv[4]);
  if (*bt < 1) {
    printf("Threshold is %d; must be >= 1\n",*bt);
    exit(1);
  }

  *form = atoi(argv[5]);
  if (!(*form == 2 || *form == 6)) {
    printf("Form Option is %d; must be either 2 or 6\n",*form);
    exit(1);
  }

  *mask_size = atoi(argv[6]);
  if (*mask_size <0) {
    printf("Mask Size is %d; must be >0\n",*mask_size);
    exit(1);
  }

  *dt = atof(argv[7]);
  if (*dt < 0.0) {
    printf("dt = %f < 0.0 must be sometinf positive\n",*dt);
    exit(1);
  }

  *tone_flag = (unsigned char)atoi(argv[8]);
  if (*tone_flag != 0 && *tone_flag != 1 && *tone_flag != 2) {
    printf("tone_flag should be either 0, 1, or 2");
    exit(1);
  }
  
}

static void error(char *name)
{
  printf("\n");
  printf("usage:  %s\n",name);
  printf("        input_img1.pgm\n");
  printf("        input_img2.pgm\n");  
  printf("        output_image.pgm\n");
  printf("        threshold\n");
  printf("        form_option\n");
  printf("              form_option = 2 => Use exp(-x^2) for weighting \n");
  printf("              form_option = 6 => Use exp(-x^6) for weighting \n");
  printf("        mask_size\n");	
  printf("        dt \n");
  printf("        tone_flag\n");
  printf("              tone_flag = 0 => do not apply tone correction \n");
  printf("              tone_flag = 1 => apply tone correction \n");

  printf("\n");
  exit(1);
}



