#include "susan_smooth.h"

float * make_gauss_filt( int mask_size,
			 float dt,
			 int x_size,
			 int y_size )
{ 

  int i, j, n_max;
  float *dp, *dpt, tmp;


  if ( (2*mask_size+1 > x_size) || (2*mask_size+1>y_size) )
    {
      printf("Mask size %d too big for image (%dx%d).\n",mask_size,x_size,y_size);
      exit(0);
    }

  n_max = (mask_size*2) + 1;
      
  if((dp = (float *)malloc(n_max*n_max*sizeof(float))) == NULL){
    fprintf(stderr," Error allocating space in make_gauss_filt");
    exit(1);
  }

  dpt    = dp;
  tmp   = -2*(dt*dt);

  for(i=-mask_size; i<=mask_size; i++)
    for(j=-mask_size; j<=mask_size; j++)
      *dpt++ =  100.0 * exp( ((float)((i*i)+(j*j))) / tmp );

  return dp;
}
      
 
void setup_brightness_lut( float **bp,
			   int thresh,
			   int form )

{
  int   k;
  float temp;

  *bp = ( float * ) malloc( 516 * sizeof(float) );
  *bp=*bp+258;

  for(k=-256;k<257;k++)
    {
      temp=((float)k)/((float)thresh);
      temp=temp*temp;
      if (form==6)
	temp=temp*temp *temp;
      *(*bp+k)= 100.0*exp(-temp);
    }
}

void susan_smooth(
  unsigned char **in1,
  float **in2,
  float ***out,
  float *bp,
  float *dp,
  int mask_size,
  int x_size,
  int y_size )

{
  
  int   x, y, i, j, centre;

  float tmp, total, area;
  float *dpt, *cp;

  /* process the non-boundary pixels */
  for(i = mask_size; i < y_size - mask_size; i++)
    for(j = mask_size; j < x_size - mask_size; j++) {
      area   = 0;
      total  = 0;
      dpt    = dp;
      centre = *(*(in1 + i) + j);
      cp     = bp + centre;     
      for(y = -mask_size; y <= mask_size; y++)
	for(x = -mask_size; x <= mask_size; x++)
	  {
            tmp = *dpt++ * *(cp - *(*(in1 + i + y) + j + x));
            area += tmp;
            total += tmp * *(*(in2 + i + y) + j + x);
	  }
      
      *(*(*out + i) + j) =  total/area;
    }

  /* take care of the boundary pixels */
  for(i = 0; i < mask_size; i++)
    for(j = mask_size; j < x_size - mask_size; j++){
      /* top */
      *(*(*out + i) + j) = *(*(*out + mask_size) + j);
      /* bottom */
      *(*(*out + i + y_size - mask_size) + j) = 
        *(*(*out + i + y_size - mask_size - 1) + j);
  }

  for(i = 0; i < y_size; i++)
    for(j = 0; j < mask_size; j++){
      /* left */
      *(*(*out + i) + j) = *(*(*out + i) + mask_size);
      /* right */
      *(*(*out + i) + j + x_size - mask_size) = 
        *(*(*out + i) + j + x_size - mask_size - 1);
  } 
}


void inv_scanner_fun ( unsigned char **gam_img,
                       float **lin_img,
                       int ht,
                       int wd,
		       unsigned char tone_flag )
{
  int         i, j;
  float      lut[256];
  float      as, bs, cs, ap, bp, cp;

  set_calib_params( tone_flag, &as, &bs, &cs, &ap, &bp, &cp );

  for (i = 0; i < 256; i++)
    lut[i] = as * pow ((double)(i * (1 / 255.0)), bs) + cs;

  for( i = 0; i < ht; i++ )
    for( j = 0; j < wd; j++ )
        *(*(lin_img + i) + j) = lut[*(*(gam_img + i) + j)];
}

void inv_printer_fun( unsigned char **out,
                      float **lin_img,
                      int ht,
                      int wd,
                      unsigned char tone_flag )

{
  int i, j;
  float tmp;
  float as, bs, cs, ap, bp, cp;

  set_calib_params(tone_flag, &as, &bs, &cs, &ap, &bp, &cp);

  for(i = 0; i < ht; i++)
    for(j = 0; j < wd; j++){

      /* apply the inv printer function */
  
      tmp = (( *(*(lin_img + i) + j) - cp) / ap);
      tmp = ((tmp > 1) ? 1 : ((tmp < 0) ? 0 : pow(tmp, 1 / bp))) * 255.;
      *(*(out + i) + j) = (unsigned char)tmp;
    }
}

void set_calib_params( unsigned char tone_flag,
                       float *as,
                       float *bs,
                       float *cs,
		       float *ap,
		       float *bp,
		       float *cp )
{
  
 switch( tone_flag )
    {

    case 0:
      *as = 1.0; /* scanner params */
      *bs = 1.0;
      *cs = 0.0;

      *ap = 1.0; /* printer params */
      *bp = 1.0;
      *cp = 0.0;
      break;

    case 1:
      *as = AS; /* scanner params */ 
      *bs = BS;
      *cs = CS;

      *ap = AP; /* printer params */
      *bp = BP;
      *cp = CP;
      break;

    case 2:
      *as = AS; /* scanner params */
      *bs = BS;
      *cs = CS;

      *ap = AS; /* printer params */
      *bp = BS;
      *cp = CS;
      break;


    default:
      printf("\nError tone_flag should be 0 or 1\n");
      printf("    0 ==> no tone correction\n ");
      printf("    1 ==> tone correction\n" );
      exit(1);
      break;

    }

}



